#!/usr/bin/env python3
from __future__ import annotations

import argparse
import logging
import os
import re
from pathlib import Path
from typing import cast

import parsy as ps
from colorama import Back, Fore, Style
from dotenv import dotenv_values
from parsy import Parser, regex, seq, string

LEVEL_UNKNOWN = 255
NEWLINE = string("\n")
SPACE = string(" ")


class Config(dict[str, str]):
    class LevelAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None) -> None:
            _ = (parser, option_string)
            values = cast(str, values)
            setattr(namespace, self.dest, format_debug_level(values.upper())[0])

    def __init__(self) -> None:
        self.dict = {k: v for k, v in dotenv_values(".env").items() if v is not None}
        parser = argparse.ArgumentParser()
        if not ("LOGFILE" in self.dict or "LOGFILE" in os.environ):
            parser.add_argument(
                "LOGFILE",
                type=Path,
            )
        parser.add_argument("-q", "--query", dest="QUERY")
        parser.add_argument(
            "-l",
            "--level",
            dest="LOGLEVEL",
            type=str,
            choices=["debug", "info", "warn", "err", "crit", "unk"],
            default=logging.INFO,
            action=self.LevelAction,
        )

        self.dict.update(parser.parse_args().__dict__)

    def __getitem__(self, key: str, /) -> str:
        try:
            return self.dict[key]
        except KeyError:
            return os.environ[key]

    def __setitem__(self, _key: str, _value: str, /) -> None:
        msg = "Config is read-only"
        raise NotImplementedError(msg)

    def __delitem__(self, _: str, /) -> None:
        msg = "Config is read-only"
        raise NotImplementedError(msg)

    def __iter__(self):
        return iter(self.dict)

    def __len__(self) -> int:
        return len(self.dict)

    def __repr__(self) -> str:
        return self.dict.__repr__()


"""
Formatting functions
"""


def colorize_status(status: str) -> str:
    if "PASS" in status:
        return Fore.GREEN + status + Style.RESET_ALL
    elif "FAIL" in status:
        return Fore.RED + status + Style.RESET_ALL
    elif "ERROR" in status:
        return Fore.BLACK + Back.RED + status + Style.RESET_ALL
    return status


def format_debug_level(level: str) -> tuple[int, str]:
    if level == "DEBUG":
        return (logging.DEBUG, level)
    elif level == "INFO":
        return (logging.INFO, level)
    elif "WARN" in level:
        return (logging.WARNING, level)
    elif "ERR" in level:
        return (logging.ERROR, level)
    elif "CRIT" in level:
        return (logging.CRITICAL, level)
    else:
        return (LEVEL_UNKNOWN, level)


def colorize_by_level(level: str, text: str) -> str:
    if level == "DEBUG":
        return Style.DIM + text + Style.RESET_ALL
    elif level == "INFO":
        return Style.BRIGHT + text + Style.RESET_ALL
    elif level == "WARNING":
        return Fore.YELLOW + text + Style.RESET_ALL
    elif level == "ERROR":
        return Fore.RED + text + Style.RESET_ALL
    elif level == "CRITICAL":
        return Style.BRIGHT + Back.RED + Fore.BLACK + text + Style.RESET_ALL
    else:
        return Fore.BLUE + text + Style.RESET_ALL
    # return level


"""
Parser defs
"""


def block_parser(text: str) -> Parser:
    return regex(rf"^.*\n^.*{text}.*$\n.*", re.MULTILINE)


timestamp: Parser = regex(r"\d{4}-\d{2}-\d{2}.\d{2}:\d{2}:\d{2}(,\d{3})?")
debug_level: Parser = regex(r"\S+")


start_block = block_parser("EASYPY JOB START")
end_block = block_parser("EASYPY JOB END")
report_block = block_parser("Easypy Report")
summary_block = block_parser("Task Result Summary")
details_block = block_parser("Task Result Details")


sections = seq(
    header=ps.any_char.until(start_block).concat() << start_block << NEWLINE,
    test_output=ps.any_char.until(end_block).concat() << end_block << NEWLINE,
    post_test=ps.any_char.until(report_block).concat() << report_block,
    report=ps.any_char.until(summary_block).concat() << summary_block,
    summary=ps.any_char.until(details_block).concat() << details_block,
    details=ps.any_char.until(NEWLINE + ps.letter).concat() << NEWLINE,
    etc=ps.any_char.until(ps.eof).concat(),
)
messsage = ps.any_char.until(ps.peek(timestamp) | ps.eof).concat()
test_output = seq(
    ts=timestamp << SPACE,
    level=debug_level.map(format_debug_level) << SPACE,
    src=string("[") >> regex(r"[A-Za-z_0-9]+") << string("] "),
    msg=messsage.map(lambda x: x.strip()),
).many()


test_details: Parser = seq(
    test_name=(
        NEWLINE.optional()
        << (SPACE * 4)
        << (string("|--") | string("`--"))
        << ps.whitespace
    )
    >> ps.any_char.until(ps.whitespace).concat()
    << ps.whitespace,
    test_results=(ps.letter.many().concat()).map(colorize_status),
)

subtask_details: Parser = seq(
    subtask_name=(string("`--") << ps.whitespace)
    >> ps.any_char.until(ps.whitespace).concat()
    << ps.whitespace,
    subtask_result=(ps.letter.many().concat()).map(colorize_status),
    tests=test_details.many(),
)

task_details: Parser = seq(
    task_number=ps.any_char.until(string(":")).concat() << string(":") << ps.whitespace,
    task_name=ps.any_char.until(ps.whitespace).concat() << ps.whitespace,
    subtasks=subtask_details.many(),
)
details: Parser = task_details.many()


"""
Output functions
"""


def output_details(log_details: list) -> None:
    for task in log_details:
        print(
            f"{task['task_number']}\t {Style.BRIGHT + task['task_name'] + Style.RESET_ALL}"
        )
        for subtask in task["subtasks"]:
            print(f"  {subtask['subtask_result']}\t{subtask['subtask_name']}")
            for test in subtask["tests"]:
                print(f" |  {test['test_results']}\t{test['test_name']}")


def output_tests(
    tests: list,
    filter_level: int = logging.NOTSET,
    search_term: str | re.Pattern | None = None,
) -> None:
    for test in tests:
        level = test["level"]
        if (search_term and not re.search(search_term, test["msg"])) or level[
            0
        ] >= filter_level:
            print(
                colorize_by_level(
                    level[1], f"{test['ts']} {level[1]} {test['src']} {test['msg']}"
                )
            )


if __name__ == "__main__":
    config = Config()
    with Path(config["LOGFILE"]).open() as f:
        log = f.read()
    parsed_sections = sections.parse(log)
    output_details(details.parse(parsed_sections["details"]))
    print("-------------\v")
    output_tests(test_output.parse(parsed_sections["test_output"]), config["LOGLEVEL"])  # noqa
